// Copyright 2017 The Johns Hopkins University Applied Physics Laboratory.
// Licensed under the MIT License. See LICENSE.txt in the project root for full license information.
/***/
/*
Calculates the geometric laplacian for both input solution and Lidar data.
Outputs:
- GLx: X-axis geometric laplacian value for input solution
- GLy: Y-axis geometric laplacian value for input solution
- GLz: Z-axis geometric laplacian value for input solution
- GLxref: X-axis geometric laplacian value for LIDAR data
- GLyref: Y-axis geometric laplacian value for LIDAR data
- GLzref: Z-axis geometric laplacian value for LIDAR data

Inputs:
- trimesh: buffer containing the vertex to cell grid correspondences
- trimeshref: buffer containing the 2D gridded X values of LIDAR data
- SZnodes: buffer containing the 2D gridded Y values of LIDAR data
- SZnodesref: buffer containing the 2D gridded Z values of LIDAR data
- Nnodes: buffer containing the 2D gridded X values of input solution
- Nnodesref: buffer containing the 2D gridded Y values of input solution
- MAXFaces: buffer containing the 2D gridded Z values of input solution
- xbuf:buffer containing the X values of input solution
- ybuf:buffer containing the Y values of input solution
- zbuf:buffer containing the Z values of input solution
- xbufref:buffer containing the X values of input solution
- ybufref:buffer containing the Y values of LIDAR data
- zbufref:buffer containing the Z values of LIDAR data


Version history: 2017-03-28	Marc Bosch	Initial version
*/
/***/
#include <math.h>
#include "mex.h"
#include <stdio.h>
#include <stdlib.h>
#include <limits>


#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{	
    if(nlhs > 6){
        mexErrMsgTxt("Function has six return value");
        return;
    }

    if(nrhs !=13){
        mexErrMsgTxt("Usage: [glx,gly,glz,glxgt,glygt,glzgt]=mexGeomLapErrorC(double *trimesh, double *trimeshref, double SZnodes, double SZnodesref, double Nnodes, double Nnodesref, double MAXFaces, double *xbuf, double *ybuf, double *zbuf, double *xbufref, double *ybufref, double *zbufref)");
        return;
    }
    const double eps = 0.01;
    /* get the image buffer*/    
    double *trimesh = (double*) mxGetPr(prhs[0]);	
    int* dims = (int*) mxGetDimensions(prhs[0]);
    int WorldDim = dims[0];//flipped since it comes transpose and row major
    int Nfaces = dims[1];

	double *trimeshref = (double*) mxGetPr(prhs[1]);	
    int* dimsref = (int*) mxGetDimensions(prhs[1]);
    int WorldDimref = dimsref[0];//flipped since it comes transpose and row major
    int Nfacesref = dimsref[1];

	double* pSZnodes = mxGetPr(prhs[2]);
 	int SZnodes = (int) *pSZnodes;

	double* pSZnodesref = mxGetPr(prhs[3]);
 	int SZnodesref = (int) *pSZnodesref;

	double* pNnodes = mxGetPr(prhs[4]);
 	int Nnodes = (int) *pNnodes;
	
	double* pNnodesref = mxGetPr(prhs[5]);
 	int Nnodesref = (int) *pNnodesref;

	double* pMAXFaces = mxGetPr(prhs[6]);
 	int MAXFaces = (int) *pMAXFaces;

	
 	int MAXFacesref = MAXFaces;

	double *xbuf = (double*) mxGetPr(prhs[7]);
	double *ybuf = (double*) mxGetPr(prhs[8]);
	double *zbuf = (double*) mxGetPr(prhs[9]);
	
	double *xbufref = (double*) mxGetPr(prhs[10]);
	double *ybufref = (double*) mxGetPr(prhs[11]);
	double *zbufref = (double*) mxGetPr(prhs[12]);

	//outputs
	
	plhs[0] = mxCreateDoubleMatrix(1, Nnodes, mxREAL);
    double *GLx = mxGetPr(plhs[0]);  
    for (int i=0; i < Nnodes; i++)
        GLx[i] = xbuf[i]; 
	plhs[1] = mxCreateDoubleMatrix(1, Nnodes, mxREAL);
    double *GLy = mxGetPr(plhs[1]);  
    for (int i=0; i < Nnodes; i++)
        GLy[i] = ybuf[i]; 
	plhs[2] = mxCreateDoubleMatrix(1, Nnodes, mxREAL);
    double *GLz = mxGetPr(plhs[2]);  
    for (int i=0; i < Nnodes; i++)
        GLz[i] = zbuf[i]; 

	plhs[3] = mxCreateDoubleMatrix(1, Nnodesref, mxREAL);
    double *GLxref = mxGetPr(plhs[3]);  
    for (int i=0; i < Nnodesref; i++)
        GLxref[i] = xbufref[i]; 
	plhs[4] = mxCreateDoubleMatrix(1, Nnodesref, mxREAL);
    double *GLyref = mxGetPr(plhs[4]);  
    for (int i=0; i < Nnodesref; i++)
        GLyref[i] = ybufref[i]; 
	plhs[5] = mxCreateDoubleMatrix(1, Nnodesref, mxREAL);
    double *GLzref = mxGetPr(plhs[5]);  
    for (int i=0; i < Nnodesref; i++)
        GLzref[i] = zbufref[i]; 
	//input mesh
	//temp variables and buffers
	int *c = (int*) malloc(SZnodes*sizeof(int));
	for (int i=0; i < SZnodes; i++)
		c[i] = 0;
	int *neighbornodes=(int*)malloc(MAXFaces*sizeof(int));
	int MAXNeigs = MAXFaces;
	int *neighbors=(int*)malloc(MAXNeigs*SZnodes*sizeof(int));
	double *sumnumerx=(double *)malloc(Nnodes*sizeof(double));
	for (int i=0; i < Nnodes; i++)
		sumnumerx[i] = 0;
	double *sumnumery=(double *)malloc(Nnodes*sizeof(double));
	for (int i=0; i < Nnodes; i++)
		sumnumery[i] = 0;
	double *sumnumerz=(double *)malloc(Nnodes*sizeof(double));
	for (int i=0; i < Nnodes; i++)
		sumnumerz[i] = 0;
	double *sumdenom=(double *)malloc(Nnodes*sizeof(double));	
	for (int i=0; i < Nnodes; i++)
		sumdenom[i] = 0;
		
	//extend xbuf,... and xbufref,... to size=SZnodes/SZnodesref
	double * xbuf_prime= (double*) malloc(SZnodes*sizeof(double));
	double * ybuf_prime= (double*) malloc(SZnodes*sizeof(double));
	double * zbuf_prime= (double*) malloc(SZnodes*sizeof(double));
	int *pair = (int*) malloc(SZnodes*sizeof(int));
	int *flagunique = (int*) malloc(SZnodes*sizeof(int));
	int *unique_nodes = (int*) malloc(Nnodes*sizeof(int));
	for(int i=0;i<Nnodes;i++)
		unique_nodes[i]=0;
	for(int i=0;i<SZnodes;i++)
	{
		xbuf_prime[i]=0;ybuf_prime[i]=0;zbuf_prime[i]=0;
		pair[i]=-1;flagunique[i]=0;
	}
	for (int k=0;k<Nfaces;k++)//loop through all faces	
		for (int j=0;j<WorldDim;j++)
		{
			flagunique[(int)trimesh[k*WorldDim+j]-1] = 1;
		}
	int count=0;
	for(int i=0;i<SZnodes;i++)
		if(flagunique[i]==1)
			unique_nodes[count++] = i;
	
	for (int k=0;k<Nnodes;k++)//loop through all faces
	{
		int nodeid=unique_nodes[k];
		xbuf_prime[nodeid]=xbuf[k];
		ybuf_prime[nodeid]=ybuf[k];
		zbuf_prime[nodeid]=zbuf[k];
		pair[nodeid] = k;
	}


	for (int k=0;k<Nfaces;k++)//loop through all faces
	{
		for (int j=0;j<WorldDim;j++)
		{	
			//store for each node all its neighbors (same face vertices)
			int nodeid=(int)trimesh[k*WorldDim+j]-1;

			for (int jj=0;jj<WorldDim;jj++)//loop through all nodes
			{
				if(jj!=j)
				{
					int neighborid=(int)trimesh[k*WorldDim+jj]-1;
					
					bool repeated=false;
					for(int kk=0;kk<c[nodeid];kk++)
						if(neighbors[nodeid*MAXNeigs+kk]==neighborid)
						{
							repeated=true;
							break;
						}
					//compute distance if first time neighbor
					if(!repeated && nodeid!=neighborid)
					{
						neighbors[nodeid*MAXNeigs+c[nodeid]] = neighborid;
						c[nodeid]++;
						double dist=sqrt((xbuf_prime[nodeid]-xbuf_prime[neighborid])*(xbuf_prime[nodeid]-xbuf_prime[neighborid])+
							(ybuf_prime[nodeid]-ybuf_prime[neighborid])*(ybuf_prime[nodeid]-ybuf_prime[neighborid])+
							(zbuf_prime[nodeid]-zbuf_prime[neighborid])*(zbuf_prime[nodeid]-zbuf_prime[neighborid]));	
						double denom = (dist==0) ? eps : 1/dist;
						sumnumerx[pair[nodeid]] += xbuf_prime[neighborid] * denom;
						sumnumery[pair[nodeid]] += ybuf_prime[neighborid] * denom;
						sumnumerz[pair[nodeid]] += zbuf_prime[neighborid] * denom;
						sumdenom[pair[nodeid]] += denom;
					}
				}
			}
		}
		
	}
	for (int i=0; i < Nnodes; i++)
	{
		sumdenom[i] = (sumdenom[i]==0) ? 1 : sumdenom[i];
		GLx[i] -= (sumnumerx[i]/sumdenom[i]);
		GLy[i] -= (sumnumery[i]/sumdenom[i]);
		GLz[i] -= (sumnumerz[i]/sumdenom[i]); 
	}
	free(xbuf_prime);
	free(ybuf_prime);
	free(zbuf_prime);
	free(sumdenom);
	free(sumnumerx);
	free(sumnumery);
	free(sumnumerz);
    free(c);
	free(pair);
	free(flagunique);
	free(unique_nodes);
	free(neighbornodes);
    free(neighbors);

	// repeat for reference/gt mesh
	//temp variables and buffers
	int *cref = (int*) malloc(SZnodesref*sizeof(int));
	for (int i=0; i < SZnodesref; i++)
		cref[i] = 0;
	int *neighbornodesref=(int*)malloc(MAXFacesref*sizeof(int));
	MAXNeigs = MAXFacesref;
	int *neighborsref=(int*)malloc(MAXNeigs*SZnodesref*sizeof(int));
	
		
	//extend xbuf,... and xbufref,... to size=SZnodes/SZnodesref
	double * xbufref_prime= (double*) malloc(SZnodesref*sizeof(double));
	double * ybufref_prime= (double*) malloc(SZnodesref*sizeof(double));
	double * zbufref_prime= (double*) malloc(SZnodesref*sizeof(double));
	int *pairref = (int*) malloc(SZnodesref*sizeof(int));
	int *flaguniqueref = (int*) malloc(SZnodesref*sizeof(int));
	int *unique_nodesref = (int*) malloc(Nnodesref*sizeof(int));
	for(int i=0;i<Nnodesref;i++)
		unique_nodesref[i]=0;
	for(int i=0;i<SZnodesref;i++)
	{
		xbufref_prime[i]=0;ybufref_prime[i]=0;zbufref_prime[i]=0;
		pairref[i]=-1;flaguniqueref[i]=0;
	}
	for (int k=0;k<Nfacesref;k++)//loop through all faces	
		for (int j=0;j<WorldDim;j++)
		{
			flaguniqueref[(int)trimeshref[k*WorldDim+j]-1] = 1;
		}
	count=0;
	for(int i=0;i<SZnodesref;i++)
		if(flaguniqueref[i]==1)
			unique_nodesref[count++] = i;
	
	for (int k=0;k<Nnodesref;k++)//loop through all faces
	{
		int nodeid=unique_nodesref[k];
		xbufref_prime[nodeid]=xbufref[k];
		ybufref_prime[nodeid]=ybufref[k];
		zbufref_prime[nodeid]=zbufref[k];
		pairref[nodeid] = k;
	}

	free(unique_nodesref);
	free(flaguniqueref);

	double *sumnumerxref=(double *)malloc(Nnodesref*sizeof(double));
	for (int i=0; i < Nnodesref; i++)
		sumnumerxref[i] = 0;
	double *sumnumeryref=(double *)malloc(Nnodesref*sizeof(double));
	for (int i=0; i < Nnodesref; i++)
		sumnumeryref[i] = 0;
	double *sumnumerzref=(double *)malloc(Nnodesref*sizeof(double));
	for (int i=0; i < Nnodesref; i++)
		sumnumerzref[i] = 0;
	double *sumdenomref=(double *)malloc(Nnodesref*sizeof(double));	
	for (int i=0; i < Nnodesref; i++)
		sumdenomref[i] = 0;

	for (int k=0;k<Nfacesref;k++)//loop through all faces
	{
		for (int j=0;j<WorldDim;j++)
		{	
			//store for each node all its neighbors (same face vertices)
			int nodeid=(int)trimeshref[k*WorldDim+j]-1;
			for (int jj=0;jj<WorldDim;jj++)//loop through all nodes
			{
				if(jj!=j)
				{
					int neighborid=(int)trimeshref[k*WorldDim+jj]-1;
					
					bool repeated=false;
					for(int kk=0;kk<cref[nodeid];kk++)
						if(neighborsref[nodeid*MAXNeigs+kk]==neighborid)
						{
							repeated=true;
							break;
						}
					//compute distance if first time neighbor
					if(!repeated && nodeid!=neighborid)
					{
						int deb=cref[nodeid];
						neighborsref[nodeid*MAXNeigs+deb] = neighborid;
						cref[nodeid]++;
						double dist=sqrt((xbufref_prime[nodeid]-xbufref_prime[neighborid])*(xbufref_prime[nodeid]-xbufref_prime[neighborid])+
							(ybufref_prime[nodeid]-ybufref_prime[neighborid])*(ybufref_prime[nodeid]-ybufref_prime[neighborid])+
							(zbufref_prime[nodeid]-zbufref_prime[neighborid])*(zbufref_prime[nodeid]-zbufref_prime[neighborid]));	
						double denom = (dist==0) ? eps : 1/dist;
						sumnumerxref[pairref[nodeid]] += xbufref_prime[neighborid] * denom;
						sumnumeryref[pairref[nodeid]] += ybufref_prime[neighborid] * denom;
						sumnumerzref[pairref[nodeid]] += zbufref_prime[neighborid] * denom;
						sumdenomref[pairref[nodeid]] += denom;
					}
				}
			}
		}
		
	}
	for (int i=0; i < Nnodesref; i++)
	{
		sumdenomref[i] = (sumdenomref[i]==0) ? 1 : sumdenomref[i];
		GLxref[i] -= (sumnumerxref[i]/sumdenomref[i]);
		GLyref[i] -= (sumnumeryref[i]/sumdenomref[i]);
		GLzref[i] -= (sumnumerzref[i]/sumdenomref[i]); 
	}
	free(xbufref_prime);
	free(ybufref_prime);
	free(zbufref_prime);
	free(sumdenomref);
	free(sumnumerxref);
	free(sumnumeryref);
	free(sumnumerzref);
    free(cref);
	free(pairref);	
    free(neighbornodesref);
    free(neighborsref);

	//// calculate error
	//double mean_GL = 0;
	//double std_GL = 0;
	//for (int i=0;i<Nnodes;i++)
	//{
	//	int corresp=(int)indexpointcolmajor[i];
	//	errorGL[i] = ((GLxref[corresp]-GLx[i])*
	//		(GLxref[corresp]-GLx[i])+
	//		(GLyref[corresp]-GLy[i])*
	//		(GLyref[corresp]-GLy[i])+
	//		(GLzref[corresp]-GLz[i])*
	//		(GLzref[corresp]-GLz[i]));
	//	mean_GL+=errorGL[i];
	//	std_GL+=(errorGL[i]*errorGL[i]);
	//}
	//double temp=mean_GL;
 //   mean_GL = mean_GL/Nnodes;
 //   std_GL =  sqrt((std_GL-temp*mean_GL)/(Nnodes-1));
	//for (int i=0;i<Nnodes;i++)
	//{
	//	validGL[i]=1;
	//	if (errorGL[i]>mean_GL+lambda*std_GL)
	//		validGL[i]=0;
	//}   



	
   
}
